#!/usr/bin/env python
# -*- coding: utf-8 -*-
################################################################################################
#
# Description :
# - lecture d'un fichier IFC pour en sortir un bilan des éléments structurels sous forme de pdf
# - comparaison entre un fichier IFC "référence" et d'autres fichiers afin d'en sortir
#   le bilan des différences sous forme de pdf
#
# Logiciels déjà testés :
# - IFC2X3 produits par [Revit d'Autodesk] puis lu par [eveBIM du CSTB]
#
################################################################################################
__author__  = "Xavier Jourdain - École normale supérieure Paris-Saclay"
__email__   = "xavier.jourdain@ens-paris-saclay.fr"
__license__ = "GNU GPL"
__date__    = "$mai 2019$"
__status__  = "Development"
################################################################################################
#
### TO DO LIST
# -récupérer la position des murs
# -identifier les murs posant problème
# -générer un fichier IFC en mettant en rouge les éléments posant pb
# -faire la fonction __str__ et __repr__ de la classe Evaluation
# -récupérer tous les types structurels à partir de IFCRELCONTAINEDINSPATIALSTRUCTURE
# -faire la fonction de récupération des dalles
# -faire la fonction de récupération des fondations
# -faire la fonction de récupération des poteaux
# -faire la fonction de récupération des poutres
# -récupérer les portes et les fenêtres et les afficher en lien avec chaque mur
# -dans le fichier généré
#  => faire un tri par niveau
# -faire en sorte que la variable elements_a_verifier serve à quelque chose !
# -faire un tri par matériaux ?
# pour la fonction read_a_line
#    => permettre la lecture de coordonnées type (float,float) ou (float,float,float) comme dans les lignes suivantes
#    line = "#6524= IFCCARTESIANPOINT((9.,-0.10000000000003,0.));"
#    line = "#6568= IFCCARTESIANPOINT((1.2,0.8));"
#    Exemples de lignes
#      line = "#6576= IFCPRODUCTDEFINITIONSHAPE($,$,(#6574));"
#      line = "#6555= IFCOPENINGELEMENT('3ynZ$kS1fD4A_nZixoQcno',#41,'M_Porte-Ouverture:0915 x 2134mm:431352:1',$,'Opening',#6553,#6548,$);"
#      line = "#6524= IFCCARTESIANPOINT((9.,-0.10000000000003,0.));"
#      line = "#6568= IFCCARTESIANPOINT((1.2,0.8));"
#      line = "#79= IFCDERIVEDUNIT((#75,#76,#77,#78),.USERDEFINED.,'Luminous Efficacy');"
################################################################################################
# Pour pouvoir générer des pdf
from fpdf import FPDF, HTMLMixin
# Pour connaître la date de création des fichiers IFC mais aussi du(des) pdf(s) généré(s)
from datetime import datetime
# Pour récupérer les fichiers IFC d'un dossier
import os
##########################################################
############ Variables d'entrée du script ################
##########################################################
# Nom du fichier IFC référence
nom_du_fichier_IFC_Reference="Batiment_Habitation_2018_Basic_FM_Handover_View.ifc"
# Dossier dans lequel se trouvent les fichiers étudiants
dossier_fichiers_etudiants="Fichiers_Etudiants"
# Sortie d'un bilan détaillé (bilan_detaille=True) ou synthétique (bilan_detaille=False)
bilan_detaille = True
# debug=True => Détails sortis dans la console pour les phases de débugages
# utile notamment si le fichier BILAN.pdf est vierge !
debug = False
# Type d'export de Revit posant problème
export_sans_quantites_IFC=['ViewDefinition [CoordinationView_V2.0]']
##########################################################
# POIDS POUR L'ÉVALUATION DES FICHIERS
# poids sur chaque élément structurel
poids_des_elements={"Unités":0.5,"Niveaux":0.1,"Fondations":0,"Murs":5,"Dalles":0,"Poutres":0,"Poteaux":0}
# poids sur chaque quantité associée aux éléments à noter
poids_des_sous_elements_Fichier={"Unité de longueur":1,"Unité de surface":1,"Unité de volume":1}
poids_des_sous_elements_Niveau={"Nombre":1,"Altitudes":4}
poids_des_sous_elements_Mur={"Nombre":1,"Surf_banches":1,"Volume":1,"Niveau":0}
poids_des_sous_elements_Dalle={"Nombre":1,"Surface":1,"Volume":1,"Niveau":0.5}
poids_des_sous_elements_Poutre={"Nombre":1}
poids_des_sous_elements_Poteau={"Nombre":1}
poids_des_sous_elements_Fondations={"Nombre":1}
poids_des_sous_elements={"Unités":poids_des_sous_elements_Fichier,"Niveaux":poids_des_sous_elements_Niveau,"Murs":poids_des_sous_elements_Mur,
                         "Dalles":poids_des_sous_elements_Dalle,"Poutres":poids_des_sous_elements_Poutre,"Poteaux":poids_des_sous_elements_Poteau,
                         "Fondations":poids_des_sous_elements_Fondations}
##########################################################
############ Constantes ##################################
##########################################################
IFC_Building_Elements =['IFCBEAM','IFCBEAMSTANDARDCASE','IFCCOLUMN','IFCCOLUMNSTANDARDCASE','IFCCHIMNEY',
                        'IFCMEMBER','IFCWALL','IFCWALLSTANDARDCASE','IFCWALLELEMENTEDCASE',
                        'IFCSLAB','IFCSLABELEMENTEDCASE','IFCSLABSTANDARDCASE', 'IFCROOF','IFCCOVERING',
                        'IFCRAMP','IFCSTAIR','IFCSTAIRFLIGHT','IFCRAMPFLIGHT',
                        'IFCPLATE','IFCCURTAINWALL','IFCRAILING','IFCFOOTING','IFCPILE',
                        'IFCDOOR','IFCDOORSTANDARDCASE','IFCWINDOW','IFCWINDOWSTANDARDCASE',
                        'IFCSHADINGDEVICE','IFCBUILDINGELEMENTPROXY']
# BEAM = poutre ; COLUMN = poteau ; CHIMNEY = cheminée
# MEMBER = ; WALL = WALLSTANDARDCASE = IFCWALLELEMENTEDCASE = mur ;
# SLAB = dalle ; ROOF = toit ; COVERING = toiture
# RAMP = rampe d'accès; STAIR = escalier ; STAIRFLIGHT = volée d'escalier; RAMPFLIGHT = volée de rampe d'accès
# PLATE = panneau de mur rideau ; CURTAINWALL = mur rideau ; RAILING = balustrade/garde-corps ; FOOTING = fondation ; PILE = pieu
# IFCSHADINGDEVICE = protection solaire ; IFCBUILDINGELEMENTPROXY = éléments non défini
# => https://bimstandards.fr/data-bim/objets-ifc/
# => http://bim.tech.fr/ifc/IfcBuildingElement.html#IfcP

# Pour introduire des tabulations et sauts de lignes dans les fichiers pdf
tab="&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"
saut_de_ligne="<p>&nbsp</p>"
##########################################################
########### Classes du script                #############
##########################################################

class Evaluation():
    def __init__(self,file_name,user_name,file_eval,storeys_eval,walls_eval,slabs_eval,beams_eval,columns_eval,footings_eval,bilan_eval):
        self.file_name          = file_name          # str : nom du fichier
        self.user_name          = user_name          # [] : nom de ou des utilisateur(s) ayant travaillé sur l'IFC
        self.elmt_eval          = {"Unités":file_eval,# {} : résultats de l'évaluation du fichier (Unités, ...)
                                   "Niveaux":storeys_eval,# {} : résultats de l'évaluation sur les niveaux (nb, altitudes)
                                   "Murs":walls_eval,# {} : résultats de l'évaluation sur les murs (nb, surf, vol, mat, niveau)
                                   "Dalles":slabs_eval,# {} : résultats de l'évaluation sur les dalles
                                   "Poutres":beams_eval,# {} : résultats de l'évaluation sur les poutres
                                   "Poteaux":columns_eval,# {} : résultats de l'évaluation sur les poteaux
                                   "Fondations":footings_eval}# {} : résultats de l'évaluation sur les fondations
        self.bilan_eval         =  bilan_eval        # {} : bilan par types d'éléments (file, storeys, walls, slabs, beams, columns, footings)


class Structure() :
    def __init__(self,file,storeys,walls,slabs,beams,columns,footings):
        self.file              = file                # {} : description du fichier (nom, logiciel, unités, ...)
        self.storeys           = storeys             # {} : dictionnaire des niveaux
        self.storeys_nb        = len(storeys)        # int: Nombre de niveaux
        self.struct_height     = 0.                  # flt: Hauteur totale du bâtiment
        self.struct_footprint  = 0.                  # flt: Surface projetée au sol du bâtiment
        self.walls             = walls               # {} : dictionnaire des murs
        self.walls_nb          = len(walls)          # int: Nombre de murs
        self.walls_surf_banche = 0.                  # flt: Surface banchee
        self.walls_net_area    = 0.                  # flt: Surface nette 
        self.walls_gross_area  = 0.                  # flt: Surface brute
        self.walls_net_volume  = 0.                  # flt: Volume net
        self.walls_gross_volume= 0.                  # flt: Volume brut
        self.slabs             = slabs               # {} : dictionnaire des dalles
        self.slabs_nb          = len(slabs)          # int: Nombre de dalles
        self.slabs_area        = 0.                  # flt: Surface de dalles
        self.slabs_volume      = 0.                  # flt: Volume de dalles
        self.beams             = beams               # {} : dictionnaire des poutres
        self.beams_nb          = len(beams)          # int: Nombre de poutres
        self.columns           = columns             # {} : dictionnaire des poteaux
        self.columns_nb        = len(columns)        # int: Nombre de poteaux
        self.footings          = footings            # {} : dictionnaire des fondations
        self.footings_nb       = len(footings)       # int: Nombre de fondations
        self.note              = 0.                  # str: Note finale pondérée suivant la variable "poids_des_elements"
    def __str__(self):
        return "Structure composée de {} éléments structurels de {} types différents".format(self.level,self.XYZ,self.length,self.volume_net)
    def __repr__(self):
        return "\nMur : "+str(self)

class File() :
    def __init__(self,file_name,date,user,software,IFC_type,units):
        self.name              = file_name           # str : nom du fichier
        self.date_of_creation  = date                # str : date de création du fichier IFC
        self.user              = user                # []  : nom(s) de(s) utilisateur(s) ayant créé le fichier
        self.software_name     = software[0]         # [] : nom()s du(des) logiciel(s) ayant créé le fichier IFC
        self.software_about    = software[1]         # [] : version(s) du(des) logiciel(s) ayant créé le fichier IFC
        self.software_export   = software[2]         # [] : type(sà d'export choisi(s) dans le(s) logiciel(s)
        self.IFC_type          = IFC_type            # str : version d'IFC
        self.unit_of_length    = units[0]            # str : unité de mesure de longueur
        self.unit_of_area      = units[1]            # str : unité de mesure de surface
        self.unit_of_volume    = units[2]            # str : unité de mesure de volume
    def __str__(self):
        str_to_return = "Fichier IFC ({}) nommé '{}'\n- créé le {} par {}\n".format(self.IFC_type,self.name,self.date_of_creation,self.user)
        str_to_return+= "- export via le logiciel Logiciel {} (version {})".format(self.software_name,self.software_about)                            
        if len(self.software_export) > 0:
            str_to_return += "(export de type '{}')".format(self.software_export)
        str_to_return +="\n - unités : [longueur={};surface={};volume={}]".format(self.unit_of_length,self.unit_of_area,self.unit_of_volume)
        str_to_return +="\n\n"
        return str_to_return
    def __repr__(self):
        return "\nDescription "+str(self)

class Storey():
    def __init__(self,storey_id,storey_guid,storey_name,storey_alt,storey_elmts,storey_str_types):
        self.id          = storey_id           # str : Id IFC (ex : #144)
        self.guid        = storey_guid         # str : Identifiant Unique id (ex:'0oPnWxjl5AKg1IId4CJ9zb')
        self.name        = storey_name         # str : Nom du niveau donné par l'utilisateur
        self.altitude    = round(float(storey_alt),10)   # flt : Altitude du niveau => arrondi à cause des 6.72000000000033 dû à l'écriture binaire !
        self.cont_struct_elmts = storey_elmts        # []  : Elements structurels du niveau (provenant de IFCRELCONTAINEDINSPATIALSTRUCTURE)
        self.cont_struct_types = storey_str_types    # []  : Types d'éléments structurels du niveau
        self.walls       = []                  # []  : guid des murs de ce niveau
        self.walls_gross_volume = 0.0          # flt : Volume brut de béton pour les murs de ce niveau
        self.walls_framework_surf = 0.0        # flt : Surface de banches pour les murs de ce niveau
        self.walls_length = 0.0                # flt : Longueur total de mur (<=> longueur de banches)
        self.columns     = []                  # []  : guid des poteaux de ce niveau
        self.columns_volume = 0.0              # flt : Volume de béton pour les poteaux de ce niveau
        self.beams       = []                  # []  : guid des poutres de ce niveau
        self.beams_volume = 0.0                # flt : Volume de béton pour les poutres de ce niveau
        self.slabs       = []                  # []  : guid des dalles de ce niveau
        self.slabs_volume = 0.0                # flt : Volume de béton pour les dalles de ce niveau
        self.footings    = []                  # []  : guid des fondations de ce niveau

        self.description = "Niveau {} - alt {}\n".format(self.name,self.altitude)
    def __str__(self):
        nb_el_str=0
        for i in range(0,len(self.cont_struct_elmts)):
            nb_el_str+=len(self.cont_struct_elmts[i])
        str_to_return = "{}({}) : Niveau {} - alt {}".format(self.id,self.guid,self.name,self.altitude)
        if nb_el_str < 0:
            str_to_return += " - Aucun élément structurel associé"
        else:
            if nb_el_str == 1:
                str_to_return += " - 1 seul élément structurel de type {} : {}".format(self.cont_struct_types,self.cont_struct_elmts)
            else:
                str_to_return += " - {} éléments structurels (types  rencontrés {} ) : {}".format(nb_el_str,self.cont_struct_types,self.cont_struct_elmts)
        total_volume = self.walls_net_volume + self.columns_net_volume + self.beams_net_volume + self.slabs_net_volume
        if total_volume > 0:
            str_to_return += " - {} m3 de béton".format(total_volume)
        str_to_return += "\n"
        return str_to_return
    def __repr__(self):
        return str(self)+"\n"

class Wall() :
    def __init__(self,wall_id,wall_guid,wall_level,wall_XYZ,wall_length,wall_width,wall_height,
                 wall_mat,wall_net_surface,wall_gross_surface,wall_net_volume,wall_gross_volume,wall_GPA):
        self.id          = wall_id                       # str : Id IFC (ex : #210)
        self.guid        = wall_guid                     # str : Identifiant Unique id (ex:'1u9PzEhHLDYwTmsndzCN8r')
        self.level       = wall_level                    # str : niveau auquel est attaché le mur
        self.XYZ         = wall_XYZ                      # str : position
        self.length      = round(wall_length,10)         # flt : longueur
        self.width       = round(wall_width,10)          # flt : largeur
        self.height      = round(wall_height,10)         # flt : hauteur
        self.material    = wall_mat                      # str : matériau
        self.GPA         = wall_GPA                      # flt : empreinte au sol
        self.surf_banches= round(wall_length*wall_height,10) # flt : surface de banches (vide pour plein)
        self.surf_brute  = round(wall_gross_surface,10)  # flt : surface brute (Revit => menuiseries déduites => même les surfaces normales au plan médian du mur comptent)
        self.surf_nette  = round(wall_net_surface,10)    # flt : surface nette (non défini dans Revit !)
        self.volume_brut = round(wall_gross_volume,10)   # flt : volume brut (Revit => ouvertures déduites)
        self.volume_net  = round(wall_net_volume,10)     # flt : volume net (non défini par Revit !)
        self.description = "Niveau {} - XYZ({}) - L={} m - Vnet={} m3\n".format(self.level,self.XYZ,self.length,self.volume_net)
    def __str__(self):
        return "Niveau {} - XYZ({}) - L={} m - V={} m3\n".format(self.level,self.XYZ,self.length,self.volume_net)
    def __repr__(self):
        return "Mur {} : ".format(self.id)+str(self)+"\n"

class Element():
   def __init__(self,el_id,el_name,el_nb_of_datas,el_datas):
        self.id            = el_id             # str : Id IFC (ex : #144)
        self.name          = el_name           # str : nom de l'objet (ex : IFCBUILDINGSTOREY)
        self.nb_of_datas   = el_nb_of_datas    # int : nombre de variables contenue dans l'objet
        self.datas         = el_datas          # []  : valeurs de chaque variable de l'objet
        self.attributes    = []                # []  : Id IFC cités par cet objet <=> Id IFC contenus dans les variables de l'objet
        self.relationships = []                # []  : Id IFC citant cet objet <=> Id IFC ayant cet objet parmi ses variables
   def __str__(self):
       return "{}= {}{}\n".format(self.id,self.name,self.datas)
   def __repr__(self):
       return "Element : "+str(self)+"\n"

# Class FPDF pour la génération de fichier pdf
class Fichier_FPDF(FPDF, HTMLMixin):
#    def header(self):  # Page HEADER => écrase la fonction par défaut (qui est vide)
#        # fpdf.image(name, x = None, y = None, w = 0, h = 0, type = '', link = '')
#        #(en haut à gauche => x=10,y=8)
#        self.image('images/Logo_ENS.png', 178, 280,21)
    def footer(self):  # Page FOOTER => écrase la fonction par défaut (qui est vide)
        # Position at 1.5 cm from bottom
        self.set_y(-15)
        # Arial italic 8
        self.set_font('Arial', 'I', 8)
        # Page number
        self.cell(0, 10, 'Page ' + str(self.page_no()) + '/{nb}' + ' - fichier généré le {}/{}/{} à {}h{}'.format(datetime.now().day,datetime.now().month,datetime.now().year,datetime.now().hour,datetime.now().minute), 0, 0, 'C')

##########################################################
########### FONCTIONS                        #############
##########################################################

# Mise en place des accents français en UTF-8 à la place du code "unicode" du fichier IFC
# À compléter pour (eveBIM \S\i = é ... tester les autres encodages)
def accents_UTF8(string):
    string = string.replace("\\X2\\00E0\\X0\\","à").replace("\\X2\\00C0\\X0\\","À")
    string = string.replace("\\X2\\00E2\\X0\\","â").replace("\\X2\\00C2\\X0\\","Â")
    string = string.replace("\\X2\\00E7\\X0\\","ç").replace("\\X2\\00C7\\X0\\","Ç")
    string = string.replace("\\X2\\00E8\\X0\\","è").replace("\\X2\\00C8\\X0\\","È")
    string = string.replace("\\X2\\00E9\\X0\\","é").replace("\\X2\\00C9\\X0\\","É").replace("\\S\\i","é")
    string = string.replace("\\X2\\00EA\\X0\\","ê").replace("\\X2\\00CA\\X0\\","Ê")
    string = string.replace("\\X2\\00EB\\X0\\","ë").replace("\\X2\\00CB\\X0\\","Ë") 
    string = string.replace("\\X2\\00EE\\X0\\","î").replace("\\X2\\00CE\\X0\\","Î")
    string = string.replace("\\X2\\00EF\\X0\\","ï").replace("\\X2\\00CF\\X0\\","Ï")
    string = string.replace("\\X2\\00F4\\X0\\","ô").replace("\\X2\\00D4\\X0\\","Ô")
    string = string.replace("\\X2\\00F9\\X0\\","ù").replace("\\X2\\00D9\\X0\\","Ù")
    string = string.replace("\\X2\\00FB\\X0\\","û").replace("\\X2\\00DB\\X0\\","Û")
    return string

# Lecture d'une ligne d'un fichier IFC puis décomposition en sous éléments
def read_a_line(line):
    el_id=line.split(sep="=")[0][0:]
    el_name=line.split(sep="=")[1].split(sep="(")[0][0:].replace(" ","")
    el_datas = line.split(sep=el_name)[1][1:-2].split(',')
    el_nb_of_datas = len(el_datas)
    for i_data in range(0,el_nb_of_datas):
        el_datas[i_data] = accents_UTF8(el_datas[i_data].strip("()"))
    return el_id,el_name,el_nb_of_datas,el_datas

def get_the_relationships_and_attributes(dico_elements_tmp):
    for element in dico_elements_tmp:
        for data in dico_elements_tmp[element].datas:
            if "#" in data:
                dico_elements_tmp[element].attributes.append(data)
                dico_elements_tmp[data].relationships.append(element)
    return dico_elements_tmp

# Lecture d'un fichier IFC et récupération de l'ensemble de ses objets = éléments
def read_a_file_and_get_all_the_elements(nom_fichier_IFC,dossier_tmp='./'):
    fichier_IFC = open(dossier_tmp+nom_fichier_IFC,'r')
    for line in fichier_IFC:
        if "DATA;" in line.upper():
            break
    dico_all_elements_tmp = {}
    for line in fichier_IFC:
        if "ENDSEC;" in line.upper():
            break
        el_id,el_name,el_nb_of_datas,el_datas = read_a_line(line)
        obj_El = Element(el_id,el_name,el_nb_of_datas,el_datas)
        dico_all_elements_tmp[obj_El.id] = obj_El
    fichier_IFC.close()
    dico_all_elements_tmp = get_the_relationships_and_attributes(dico_all_elements_tmp)
    if debug:
        print(" - {} éléments récupérés dans le fichier".format(len(dico_all_elements_tmp)))
    return dico_all_elements_tmp

# Lecture d'un fichier IFC et récupération des informations de base du fichier
def get_file_description(nom_fichier_IFC,dico_file_description_tmp,dossier_tmp="./"):
    fichier_IFC = open(dossier_tmp+nom_fichier_IFC,'r')
    dico_fichiers_tmp = {}
    software=[[],[],[]]
    units=["","",""]
    user=[]
    date=""
    for line in fichier_IFC:
        if "FILE_DESCRIPTION" in line.upper():
            # pour REVIT => type d'export IFC qui a été choisi dans le menu d'export IFC de REVIT
            software[2].append(line.split(sep="(('")[1].split(sep="')")[0])
            if software[2][-1] in export_sans_quantites_IFC :
                print("\n!!! ATTENTION ! AUCUNE QUANTITÉ ASSOCIÉE AUX ÉLÉMENTS AVEC L'EXPORT CHOISI ({}) !!!\n".format(software[2]))
        if "FILE_SCHEMA" in line.upper():
            # Type d'IFC => IFC2x3 ou IFC4
            IFC_type = line.split(sep="(('")[1].split(sep="'))")[0]
        if "DATA;" in line.upper():
            break
    fichier_IFC.close()
    for obj_El in dico_file_description_tmp:
        if dico_file_description_tmp[obj_El].name == "IFCAPPLICATION":
            # récupération des infos sur le logiciel ayant fait l'export
            # exemple : #5= IFCAPPLICATION(#1,'2019','Autodesk Revit 2019 (FRA)','Revit');
            software[0].append(dico_file_description_tmp[obj_El].datas[3])
            software[1].append(dico_file_description_tmp[obj_El].datas[1])
        if dico_file_description_tmp[obj_El].name == "IFCPERSON":
            # récupération du nom de l'utilisateur qui a créé le fichier IFC
            user.append(dico_file_description_tmp[obj_El].datas[2])
        if dico_file_description_tmp[obj_El].name == "IFCOWNERHISTORY":
            # récupération de la date
            # date = secondes écoulées depuis le début de l'année 1970
            date = datetime.utcfromtimestamp(int(dico_file_description_tmp[obj_El].datas[7])).strftime('%Y-%m-%d %H:%M:%S')
        if dico_file_description_tmp[obj_El].name == "IFCSIUNIT":
            # récupération des unités nécessaire pour la comparaison
            if "LENGTHUNIT" in dico_file_description_tmp[obj_El].datas[1]:
                units[0] = dico_file_description_tmp[obj_El].datas[3]
            if "AREAUNIT" in dico_file_description_tmp[obj_El].datas[1]:
                units[1] = dico_file_description_tmp[obj_El].datas[3]
            if "VOLUMEUNIT" in dico_file_description_tmp[obj_El].datas[1]:
                units[2] = dico_file_description_tmp[obj_El].datas[3]
    dico_fichiers_tmp[nom_fichier_IFC] = File(nom_fichier_IFC,date,user,software,IFC_type,units)
    return dico_fichiers_tmp

# Récupération des objets de type TYPE_TMP parmi un ensemble d'objets
def get_specific_elements(dico_all_elements_tmp,type_tmp):
#   dico_elements_tmp = dictionnaire contenant des objets
#   type_tmp = tableau contenant des chaines de caractères
#          par exemple : ['IFCWALL'] ou ['IFCSLAB','IFCWALLSTANDARDCASE']
    dico_tmp = {}
    for obj_tmp in dico_all_elements_tmp:
        if dico_all_elements_tmp[obj_tmp].name.upper() in type_tmp:
            dico_tmp[dico_all_elements_tmp[obj_tmp].id]=dico_all_elements_tmp[obj_tmp]
    if debug:
        print(" - {} éléments récupérés de types {}".format(len(dico_tmp),type_tmp))
    return dico_tmp

# Récupération des niveaux et de leurs principales caractéristiques
def get_storeys(dico_storeys_tmp,dico_elmts_by_storey,dico_all_elements_tmp):
    dico_tmp = {}
    for id_storey_tmp in dico_storeys_tmp:
        storey_id = dico_storeys_tmp[id_storey_tmp].id
        storey_guid = dico_storeys_tmp[id_storey_tmp].datas[0]
        storey_name = dico_storeys_tmp[id_storey_tmp].datas[2]
        storey_alt = dico_storeys_tmp[id_storey_tmp].datas[9]
        id_storey_elmts = ""
        for id_tmp in dico_elmts_by_storey:
            if dico_elmts_by_storey[id_tmp].datas[dico_elmts_by_storey[id_tmp].nb_of_datas-1] == id_storey_tmp:
                id_storey_elmts = id_tmp
                break
        storey_elmts = []
        storey_str_types = []
        for i_tmp in range(4 , dico_elmts_by_storey[id_storey_elmts].nb_of_datas - 1):
#            storey_elmts.append(dico_elmts_by_storey[id_storey_elmts].datas[i_tmp])
            if dico_all_elements_tmp[dico_elmts_by_storey[id_storey_elmts].datas[i_tmp]].name not in storey_str_types:
                storey_str_types.append(dico_all_elements_tmp[dico_elmts_by_storey[id_storey_elmts].datas[i_tmp]].name)
        for i_tmp in range(0,len(storey_str_types)):
            storey_elmts.append([])
        for i_tmp in range(4 , dico_elmts_by_storey[id_storey_elmts].nb_of_datas - 1):
            for i_type_tmp in range(0,len(storey_str_types)):
                if dico_all_elements_tmp[dico_elmts_by_storey[id_storey_elmts].datas[i_tmp]].name == storey_str_types[i_type_tmp]:
                    storey_elmts[i_type_tmp].append(dico_elmts_by_storey[id_storey_elmts].datas[i_tmp])
        obj_Storey = Storey(storey_id,storey_guid,storey_name,storey_alt,storey_elmts,storey_str_types)
        dico_tmp[obj_Storey.id] = obj_Storey
    return dico_tmp

# Récupération des murs et de leurs principales caractéristiques
def get_walls_elements(dico_all_elements_tmp,dico_storeys_tmp):
#   dico_all_elements_tmp = dictionnaire contenant tous les éléments du fichier IFC
    dico_tmp = {}
    for elmt_tmp in dico_all_elements_tmp:
        if dico_all_elements_tmp[elmt_tmp].name.upper() in ['IFCWALL','IFCWALLSTANDARDCASE']:
            dico_tmp[dico_all_elements_tmp[elmt_tmp].id] = dico_all_elements_tmp[elmt_tmp]
    dico_walls_tmp = {}
    for elmt_tmp in dico_tmp:
        wall_XYZ = ""
        wall_mat = ""
        wall_height = 0.0
        wall_length = 0.0
        wall_width = 0.0
        wall_net_surface = 0.0
        wall_gross_surface = 0.0
        wall_net_volume = 0.0
        wall_gross_volume = 0.0
        wall_GPA = 0.0
        wall_id = dico_all_elements_tmp[elmt_tmp].id
        wall_guid = dico_all_elements_tmp[elmt_tmp].datas[0]
#        print("{}".format(dico_all_elements_tmp[elmt_tmp].attributes))
#        print("{}\n\n".format(dico_all_elements_tmp[elmt_tmp].relationships))
        for elmt_rel in dico_all_elements_tmp[elmt_tmp].relationships:
            el_qtty_id = dico_all_elements_tmp[elmt_rel].datas[dico_all_elements_tmp[elmt_rel].nb_of_datas-1]
            if ((dico_all_elements_tmp[elmt_rel].name == "IFCRELDEFINESBYPROPERTIES") and
                dico_all_elements_tmp[el_qtty_id].name == "IFCELEMENTQUANTITY"):
                    for i_tmp in range(5,11):
                        str_tmp = dico_all_elements_tmp[dico_all_elements_tmp[el_qtty_id].datas[i_tmp]].datas[3]
                        if dico_all_elements_tmp[dico_all_elements_tmp[el_qtty_id].datas[i_tmp]].datas[0]=="'Height'":
                            wall_height = float(str_tmp)
                        if dico_all_elements_tmp[dico_all_elements_tmp[el_qtty_id].datas[i_tmp]].datas[0]=="'Length'":
                            wall_length = float(str_tmp)
                        if dico_all_elements_tmp[dico_all_elements_tmp[el_qtty_id].datas[i_tmp]].datas[0]=="'Width'":
                            wall_width = float(str_tmp)
                        if dico_all_elements_tmp[dico_all_elements_tmp[el_qtty_id].datas[i_tmp]].datas[0]=="'GrossFootprintArea'":
                            wall_GPA = float(str_tmp)
                        if dico_all_elements_tmp[dico_all_elements_tmp[el_qtty_id].datas[i_tmp]].datas[0]=="'GrossSideArea'":
                            wall_gross_surface = float(str_tmp)
                        if dico_all_elements_tmp[dico_all_elements_tmp[el_qtty_id].datas[i_tmp]].datas[0]=="'GrossVolume'":
                            wall_gross_volume = float(str_tmp)
        ####################################################################################
        # AJOUT POUR REVIT 2019 QUI DEFINIT MAL LES QUANTITES !!!                          #
                    if wall_gross_volume   <   wall_height * wall_length * wall_width :    #
                        wall_net_volume = wall_gross_volume                                #
                        wall_gross_volume = wall_height * wall_length * wall_width         #
        # AJOUT POUR REVIT 2019 QUI DEFINIT MAL LES QUANTITES !!!                          #
        #################################################################################### 
            if (dico_all_elements_tmp[elmt_rel].name == "IFCRELASSOCIATESMATERIAL"):
                id_tmp = dico_all_elements_tmp[elmt_rel].datas[5]
                if dico_all_elements_tmp[id_tmp].name == "IFCMATERIALLAYERSETUSAGE":
                    id_tmp = dico_all_elements_tmp[id_tmp].datas[0]
                    if dico_all_elements_tmp[id_tmp].name == "IFCMATERIALLAYERSET":
                        id_tmp = dico_all_elements_tmp[id_tmp].datas[0]
                        if dico_all_elements_tmp[id_tmp].name == "IFCMATERIALLAYER":
                            id_tmp = dico_all_elements_tmp[id_tmp].datas[0]
                        if dico_all_elements_tmp[id_tmp].name == "IFCMATERIAL":
                            wall_mat = dico_all_elements_tmp[id_tmp].datas[0]
                if wall_mat == "":
                    print("\nProblème d'attribution de matériau pour le mur {} !\n".format(elmt_tmp))
        wall_level = ""
        for obj_tmp in dico_storeys_tmp:
            for elmt_struct_tmp in dico_storeys_tmp[obj_tmp].cont_struct_elmts:
                if elmt_tmp in elmt_struct_tmp:
                    wall_level = dico_storeys_tmp[obj_tmp].name+"({})".format(dico_storeys_tmp[obj_tmp].altitude)
                    dico_storeys_tmp[obj_tmp].walls_gross_volume += round(wall_gross_volume,10)
                    dico_storeys_tmp[obj_tmp].walls_framework_surf += round(wall_length*wall_height,10)
                    dico_storeys_tmp[obj_tmp].walls_length += round(wall_length,10)
                    dico_storeys_tmp[obj_tmp].walls.append(wall_guid)
                    break
        if wall_level == "":
            print("\n\n!!! mur {} attribué à aucun niveau !!!\n\n".format(elmt_tmp))
        obj_Wall = Wall(wall_id,wall_guid,wall_level,wall_XYZ,wall_length,wall_width,wall_height,wall_mat,
                        wall_net_surface,wall_gross_surface,wall_net_volume,wall_gross_volume,wall_GPA)
        dico_walls_tmp[obj_Wall.guid] = obj_Wall   
    return (dico_walls_tmp,dico_storeys_tmp)

# Récupération des dalles et de leurs principales caractéristiques
def get_slabs_elements(dico_all_elements_tmp):
#   dico_all_elements_tmp = dictionnaire contenant tous les éléments du fichier IFC
    dico_tmp = {}
    for elmt_tmp in dico_all_elements_tmp:
        if dico_all_elements_tmp[elmt_tmp].name.upper() in ['IFCSLAB','IFCWALLSTANDARDCASE']:
            dico_tmp[dico_all_elements_tmp[elmt_tmp].id] = dico_all_elements_tmp[elmt_tmp]
    dico_slabs_tmp = {}
    return dico_slabs_tmp

# Création d'un objet "Structure" à partir de la lecture des différents éléments d'un fichier IFC
def get_structure_from_IFC(nom_du_fichier_IFC_tmp,dossier_tmp = './'):
    # Récupération de l'ensemble des éléments du fichier IFC
    dico_all_elements={}
    dico_all_elements = read_a_file_and_get_all_the_elements(nom_du_fichier_IFC_tmp,dossier_tmp)    
    # Récupération des informations clés du fichier IFC
    dico_infos_fichier ={}
    dico_infos_fichier = get_file_description(nom_du_fichier_IFC_tmp,get_specific_elements(dico_all_elements,['IFCAPPLICATION','IFCPERSON','IFCOWNERHISTORY','IFCSIUNIT']),dossier_tmp)    
    # Récupération des niveaux du bâtiment
    dico_storeys={}
    dico_storeys = get_storeys(get_specific_elements(dico_all_elements,['IFCBUILDINGSTOREY']),
                               get_specific_elements(dico_all_elements,['IFCRELCONTAINEDINSPATIALSTRUCTURE']),
                               dico_all_elements)    
    # Récupération de tous les éléments de typeMUR
    dico_walls = {}
    (dico_walls,dico_storeys) = get_walls_elements(dico_all_elements,dico_storeys)  
    # Récupération de tous les éléments de type DALLE
    dico_slabs = {}
    dico_slabs = get_slabs_elements(dico_all_elements)    
    # Récupération de tous les éléments de type POUTRE
    dico_beams = {}
    #dico_beams = get_beams_elements(dico_all_elements)    
    # Récupération de tous les éléments de type POTEAU
    dico_columns = {}
    #dico_columns = get_slabs_elements(dico_all_elements)    
    # Récupération de tous les éléments de type FONDATION
    dico_footings = {}
    #dico_footings = get_footings_elements(dico_all_elements)
    # Création dun objet structure récapitulant tous les éléments enregistrés
    structure = Structure(dico_infos_fichier[nom_du_fichier_IFC_tmp],dico_storeys,dico_walls,dico_slabs,dico_beams,dico_columns,dico_footings)
    # Remplissage des données globales de la structure
    for wall_i in dico_walls:
        structure.walls_gross_area   += dico_walls[wall_i].surf_brute
        structure.walls_gross_volume += dico_walls[wall_i].volume_brut
        structure.walls_surf_banche  += dico_walls[wall_i].surf_banches
        structure.walls_net_area     += dico_walls[wall_i].surf_nette
        structure.walls_net_volume   += dico_walls[wall_i].volume_net
    return structure

# Génération d'un fichier pdf BILAN à partir d'un dictionnaire d'éléments structurels d'un fichier
def generate_pdf_from_one_structure(dico_structure_tmp):
    pdf_bilan_IFC = Fichier_FPDF()
    pdf_bilan_IFC.alias_nb_pages()  # pour avoir le nombre total de pages dans {nb}   
    bilan_html='<H1 align="center"> Bilan du fichier IFC</H1>'
    bilan_html+='<H3 align="center"> {}</H3>'.format(dico_structure_tmp.file.name)
    # Sortie des informations générales du fichier
    numero_de_chapitre=1
    bilan_html+='<H2 align="left"> {} - Informations générales</H2>'.format(numero_de_chapitre)
    if len(dico_structure_tmp.file.user) == 1:
        bilan_html+="<p><b>Nom de l'utilisateur</b> : {}</p>".format(dico_structure_tmp.file.user[0]) 
    else:
        bilan_html+="<p><b>Nom des utilisateurs successifs</b> : {}".format(dico_structure_tmp.file.user[0])
        for i_user in range(1,len(dico_structure_tmp.file.user)):
            bilan_html+=";{}".format(dico_structure_tmp.file.user[i_user])
        bilan_html+="</p>"
    bilan_html+="<p><b>Date de création du fichier</b> : {}</p>".format(dico_structure_tmp.file.date_of_creation)
    if len(dico_structure_tmp.file.software_name) == 1:
        bilan_html+="<p><b>Logiciel utilisé</b> : {} version {}".format(dico_structure_tmp.file.software_name[0],dico_structure_tmp.file.software_about[0])
    else:
        bilan_html+="<p><b>Logiciels successifs utilisés</b> : {} version {}".format(dico_structure_tmp.file.software_name[0],dico_structure_tmp.file.software_about[0])
        for i_software in range(1,len(dico_structure_tmp.file.software_name)):
            bilan_html+=" ; {} version {}".format(dico_structure_tmp.file.software_name[i_software],dico_structure_tmp.file.software_about[i_software])
        bilan_html+="</p>"  
    if len(dico_structure_tmp.file.software_export) > 0:
        if dico_structure_tmp.file.software_export in export_sans_quantites_IFC:
            bilan_html+="<p><b>!!! ATTENTION ! AUCUNE QUANTITÉ ASSOCIÉE AUX ÉLÉMENTS AVEC L'EXPORT CHOISI ({}) !!!</b></p>".format(dico_structure_tmp.file.software_export)
        else:
            bilan_html+=" <i>(type d'export : '{}')</i>".format(dico_structure_tmp.file.software_export)
    bilan_html+="</p>"
    bilan_html+="<p><b>Format IFC</b> : {}</p>".format(dico_structure_tmp.file.IFC_type)
    bilan_html+="<p><b>Unité de longueur</b> : {}</p>".format(dico_structure_tmp.file.unit_of_length)
    bilan_html+="<p><b>Unité de surface</b> : {}</p>".format(dico_structure_tmp.file.unit_of_area)    
    bilan_html+="<p><b>Unité de volume</b> : {}</p>".format(dico_structure_tmp.file.unit_of_volume)    
    bilan_html+=saut_de_ligne
    numero_de_chapitre+=1
    bilan_html+='<H2 align="left"> {} - Bilan par niveau</H2>'.format(numero_de_chapitre)
    if len(dico_structure_tmp.storeys) > 0:
        numero_de_sous_chapitre = 1
        for storey_i in dico_structure_tmp.storeys :         
            bilan_html+='<H3 align="left"> {}.{} - Niveau {} (altitude {})</H3>'.format(numero_de_chapitre,numero_de_sous_chapitre,dico_structure_tmp.storeys[storey_i].name,dico_structure_tmp.storeys[storey_i].altitude)
            bilan_html+="<p><b>Nom</b> : {}</p>".format(dico_structure_tmp.storeys[storey_i].name)
            bilan_html+="<p><b>Altitude</b> : {}</p>".format(dico_structure_tmp.storeys[storey_i].altitude)
            # BILAN SUR LES FONDATIONS DU NIVEAU
            # TO DO
            # BILAN SUR LES MURS DU NIVEAU
            if len(dico_structure_tmp.storeys[storey_i].walls) > 0 :
                bilan_html+="<p><b>Murs</b> :</p>"
                vol_net_total_murs = 0.
                surf_banches_totale_murs = 0.
                numero_du_mur = 1
                for id_walls in dico_structure_tmp.storeys[storey_i].walls:                    
                    if bilan_detaille:
                        bilan_html+="<p>"+tab+"M{} : L = {} ; Ép = {} ; H = {} ; Mat = {} ; S_banches = {} ; V = {}</p>".format(numero_du_mur,dico_structure_tmp.walls[id_walls].length,dico_structure_tmp.walls[id_walls].width,dico_structure_tmp.walls[id_walls].height,dico_structure_tmp.walls[id_walls].material,dico_structure_tmp.walls[id_walls].surf_banches,dico_structure_tmp.walls[id_walls].volume_brut)
                    numero_du_mur += 1
                    vol_net_total_murs+=dico_structure_tmp.walls[id_walls].volume_net
                    surf_banches_totale_murs+=dico_structure_tmp.walls[id_walls].surf_brute
                bilan_html+="<p>"+tab+"<b>=> Nombre de murs </b>: {}</p>".format(len(dico_structure_tmp.storeys[storey_i].walls))
                bilan_html+="<p>"+tab+"<b>=> Volume total de béton pour les murs </b>: {}</p>".format(round(vol_net_total_murs,10))
                bilan_html+="<p>"+tab+"<b>=> Surface totale de banches pour ces murs </b>: {}</p>".format(round(surf_banches_totale_murs,10))
            # BILAN SUR LES POTEAUX DU NIVEAU
            # TO DO
            # BILAN SUR LES POUTRES DU NIVEAU
            # TO DO
            # BILAN SUR LES DALLES DU NIVEAU
            # TO DO
            # FIN DU BILAN DU NIVEAU
            numero_de_sous_chapitre += 1
            bilan_html += saut_de_ligne + saut_de_ligne
    else:
        bilan_html+="Aucun niveau n'est définit !!!\n"
#    for dico_tmp in dico_structure:
#        bilan_html+='<H2 align="left"> {} - {}</H2>\n'.format(num_tmp,dico_tmp)
#        for elmt_tmp in dico_structure[dico_tmp]:
#            bilan_html+="<p>{}</p>".format(dico_structure[dico_tmp][elmt_tmp].description)
#        num_tmp+=1   
    pdf_bilan_IFC.add_page()
    pdf_bilan_IFC.write_html(bilan_html)
    pdf_bilan_IFC.output('Bilan_du_fichier_IFC_par_type_delements_{}.pdf'.format(dico_structure_tmp.file.name),'F')

def relative_percentage(student_tmp,ref_tmp):
    if student_tmp == ref_tmp:
        relative_percentage_tmp = 100.
    elif abs(ref_tmp) > 0:
        relative_percentage_tmp = round(100.-100.*abs((student_tmp-ref_tmp)/ref_tmp),1)
    else:
        relative_percentage_tmp = 0.
    return relative_percentage_tmp

def print_comparison_bar(perc_tmp,bar_size_tmp=10):
    str_bar_to_return = " ["    
    for i in range(0,bar_size_tmp):
        if i <= int(round(perc_tmp/100*bar_size_tmp,0)):
            str_bar_to_return+="+"
        else:
            str_bar_to_return+="-"
    str_bar_to_return+="]"
    return str_bar_to_return

def print_comparison_result(student_tmp,ref_tmp=-1.,unit_tmp=""):
    if(ref_tmp >= 0.):
        percentage_tmp = relative_percentage(student_tmp,ref_tmp)
        str_to_return = " {} {} (VS réf = {} {} => {}%)".format(student_tmp,unit_tmp,ref_tmp,unit_tmp,percentage_tmp)
        str_to_return += print_comparison_bar(percentage_tmp,10)
    else:
        str_to_return  = " {}".format(student_tmp)
    return str_to_return

def print_encouragement(percentage_tmp):
    if percentage_tmp < 50.:
        str_to_return = " => besoin de reprendre la modélisation !"
    elif percentage_tmp < 75.:
        str_to_return = " => c'est encourageant !"
    elif percentage_tmp < 100.:
        str_to_return = " => presque parfait !"
    elif percentage_tmp == 100.:
        str_to_return = " => Félicitation !"
    return str_to_return


# Génération d'un fichier pdf COMPARAISON entre un fichier IFC élève et un fichier IFC référence
def generate_pdf_for_comparison(dico_structure_ref_tmp,dico_structure_student_tmp,dico_evaluation_student_tmp):
    diff_percentage = []
    pdf_bilan_IFC = Fichier_FPDF()
    pdf_bilan_IFC.alias_nb_pages()  # pour avoir le nombre total de pages dans {nb}   
    bilan_html='<H1 align="center"> Analyse comparative de deux fichiers IFC</H1>'
    bilan_html+='<H3 align="center"> Fichier analysé : {}</H3>'.format(dico_structure_student_tmp.file.name)
    # Sortie des informations générales du fichier
    numero_de_chapitre=1
    bilan_html+='<H2 align="left"> {} - Informations générales</H2>'.format(numero_de_chapitre)
    if len(dico_structure_student_tmp.file.user) == 1:
        bilan_html+="<p><b>Nom de l'utilisateur</b> : {}</p>".format(dico_structure_student_tmp.file.user[0]) 
    else:
        bilan_html+="<p><b>Nom des utilisateurs successifs</b> : {}".format(dico_structure_student_tmp.file.user[0])
        for i_user in range(1,len(dico_structure_student_tmp.file.user)):
            bilan_html+=" ; {}".format(dico_structure_student_tmp.file.user[i_user])
        bilan_html+="</p>"  
    bilan_html+="<p><b>Date de création du fichier</b> : {}</p>".format(dico_structure_student_tmp.file.date_of_creation)
    if len(dico_structure_student_tmp.file.software_name) == 1:
        bilan_html+="<p><b>Logiciel utilisé</b> : {} version {}</p>".format(dico_structure_student_tmp.file.software_name[0],dico_structure_student_tmp.file.software_about[0])
    else:
        bilan_html+="<p><b>Logiciels successifs utilisés</b> : {} version {}".format(dico_structure_student_tmp.file.software_name[0],dico_structure_student_tmp.file.software_about[0])
        for i_software in range(1,len(dico_structure_student_tmp.file.software_name)):
            bilan_html+=" ; {} version {}".format(dico_structure_student_tmp.file.software_name[i_software],dico_structure_student_tmp.file.software_about[i_software])
        bilan_html+="</p>"  
    if len(dico_structure_student_tmp.file.software_export) > 0:
        bilan_html+="<p>" 
        for i_export in range(0,len(dico_structure_student_tmp.file.software_export)):
            if dico_structure_student_tmp.file.software_export[i_export] in export_sans_quantites_IFC:
                bilan_html+="<p><b>!!! ATTENTION ! AUCUNE QUANTITÉ ASSOCIÉE AUX ÉLÉMENTS AVEC L'EXPORT CHOISI ({}) !!!</b></p>".format(dico_structure_student_tmp.file.software_export)
            else:
                if len(dico_structure_student_tmp.file.software_export) == 1 :
                    bilan_html+=" <i>(type d'export du logiciel: '{}')</i>".format(dico_structure_student_tmp.file.software_export[i_export])
                else:
                    bilan_html+=" <i>(type d'export logiciel {} : '{}')</i>".format((i_export+1),dico_structure_student_tmp.file.software_export[i_export])
        bilan_html+="</p>"
    bilan_html+="</p>"
    bilan_html+="<p><b>Format IFC</b> : {}</p>".format(dico_structure_student_tmp.file.IFC_type)
    bilan_html+="<p><b>Unité de longueur</b> : {}</p>".format(dico_structure_student_tmp.file.unit_of_length)
    bilan_html+="<p><b>Unité de surface</b> : {}</p>".format(dico_structure_student_tmp.file.unit_of_area)    
    bilan_html+="<p><b>Unité de volume</b> : {}</p>".format(dico_structure_student_tmp.file.unit_of_volume)
    bilan_html+="<p><b>Évaluation sur le fichier</b> :</p>"
    for i_eval in dico_evaluation_student_tmp.elmt_eval["Unités"]:
        bilan_html+="<p>"+tab+"{} : {}% {}</p>".format(i_eval,dico_evaluation_student_tmp.elmt_eval["Unités"][i_eval],print_comparison_bar(dico_evaluation_student_tmp.elmt_eval["Unités"][i_eval],5))
    bilan_html+=saut_de_ligne
    numero_de_chapitre+=1
    bilan_html+='<H2 align="left"> {} - Analyse par niveau</H2>'.format(numero_de_chapitre)
    if len(dico_structure_student_tmp.storeys) > 0:
        bilan_html+="<p><b>Nombre de niveaux </b>:"
        bilan_html+=print_comparison_result(len(dico_structure_student_tmp.storeys),len(dico_structure_ref_tmp.storeys))
        bilan_html+=saut_de_ligne
        numero_de_sous_chapitre = 1
        for storey_i in dico_structure_student_tmp.storeys : 
            for storey_ref_tmp in dico_structure_ref_tmp.storeys :
                if round(abs(dico_structure_student_tmp.storeys[storey_i].altitude-dico_structure_ref_tmp.storeys[storey_ref_tmp].altitude),10) < 0.00000001:
                    storey_ref_i = storey_ref_tmp
            diff_percentage.append(["Niveau {} (altitude {}) :\n".format(dico_structure_student_tmp.storeys[storey_i].name,dico_structure_student_tmp.storeys[storey_i].altitude)])
            bilan_html+='<H3 align="left"> {}.{} - Niveau {} (altitude {})</H3>'.format(numero_de_chapitre,numero_de_sous_chapitre,dico_structure_student_tmp.storeys[storey_i].name,dico_structure_student_tmp.storeys[storey_i].altitude)
            bilan_html+="<p><b>Nom</b> : {}</p>".format(dico_structure_student_tmp.storeys[storey_i].name)
            bilan_html+="<p><b>Altitude</b> : {}</p>".format(dico_structure_student_tmp.storeys[storey_i].altitude)
            # BILAN SUR LES FONDATIONS DU NIVEAU
            # TO DO
            # BILAN SUR LES MURS DU NIVEAU
            if len(dico_structure_student_tmp.storeys[storey_i].walls) > 0 :
                bilan_html+="<p><b>Murs</b> :</p>"
                diff_perc_walls = []
                student_vol_brut_total_murs = 0.
                student_surf_banches_totale_murs = 0.
                student_longueur_totale_murs = 0.
                numero_du_mur = 1
                for id_walls in dico_structure_student_tmp.storeys[storey_i].walls:                    
                    if bilan_detaille:
                        bilan_html+="<p>"+tab+"M{} : L = {} ; Ép = {} ; H = {} ; Mat = {} ; S_banches = {} ; V = {}</p>".format(numero_du_mur,dico_structure_student_tmp.walls[id_walls].length,dico_structure_student_tmp.walls[id_walls].width,dico_structure_student_tmp.walls[id_walls].height,dico_structure_student_tmp.walls[id_walls].material,dico_structure_student_tmp.walls[id_walls].surf_banches,dico_structure_student_tmp.walls[id_walls].volume_brut)
                    numero_du_mur += 1
                    student_vol_brut_total_murs     += dico_structure_student_tmp.walls[id_walls].volume_brut
                    student_surf_banches_totale_murs += dico_structure_student_tmp.walls[id_walls].surf_banches
                    student_longueur_totale_murs +=dico_structure_student_tmp.walls[id_walls].length
                ref_vol_brut_total_murs = 0.
                ref_surf_banches_totale_murs = 0.
                ref_longueur_totale_murs = 0.
                for id_walls in dico_structure_ref_tmp.storeys[storey_ref_i].walls:                    
                    ref_vol_brut_total_murs         += dico_structure_ref_tmp.walls[id_walls].volume_brut
                    ref_surf_banches_totale_murs   += dico_structure_ref_tmp.walls[id_walls].surf_banches
                    ref_longueur_totale_murs       += dico_structure_ref_tmp.walls[id_walls].length
                student=len(dico_structure_student_tmp.storeys[storey_i].walls)
                ref=len(dico_structure_ref_tmp.storeys[storey_ref_i].walls)
                diff_perc_walls.append(relative_percentage(student,ref))
                bilan_html+="<p>"+tab+"<b>=> Nombre de murs </b>:"+print_comparison_result(student,ref)+"</p>"
                student=round(student_vol_brut_total_murs,10)
                ref=round(ref_vol_brut_total_murs,10)
                diff_perc_walls.append(relative_percentage(student,ref))
                bilan_html+="<p>"+tab+"<b>=> Volume total de béton pour les murs </b>:"+print_comparison_result(student,ref)+"</p>"
                student=round(student_surf_banches_totale_murs,10)
                ref=round(ref_surf_banches_totale_murs,10)
                diff_perc_walls.append(relative_percentage(student,ref))
                bilan_html+="<p>"+tab+"<b>=> Surface totale de banches pour ces murs </b>:"+print_comparison_result(student,ref)+"</p>"
                student=round(student_longueur_totale_murs,10)
                ref=round(ref_longueur_totale_murs,10)
                diff_perc_walls.append(relative_percentage(student,ref))
                bilan_html+="<p>"+tab+"<b>=> Longueur totale de murs (~longueur de banches)</b>:"+print_comparison_result(student,ref)+"</p>"
                diff_perc_walls_TOTAL=0.
                for i in range (0,len(diff_perc_walls)):
                    diff_perc_walls_TOTAL += diff_perc_walls[i]/len(diff_perc_walls)
                diff_percentage[numero_de_sous_chapitre-1].append("<b>Murs </b>: {} %".format(round(diff_perc_walls_TOTAL,1)))
                diff_percentage[numero_de_sous_chapitre-1][-1]+= print_comparison_bar(diff_perc_walls_TOTAL,10)
            # BILAN SUR LES POTEAUX DU NIVEAU
            # TO DO
            # BILAN SUR LES POUTRES DU NIVEAU
            # TO DO  
            # BILAN SUR LES DALLES DU NIVEAU
            # TO DO
            # FIN DU BILAN DU NIVEAU
            numero_de_sous_chapitre += 1
            bilan_html += saut_de_ligne
        if len(diff_percentage) > 0:            
            bilan_html+='<H3 align="left"> {}.{} - Résumé du bilan par niveau </H3>'.format(numero_de_chapitre,numero_de_sous_chapitre)
            for level_tmp in range(0,len(diff_percentage)):
                bilan_html+="<p>{}</p>".format(diff_percentage[level_tmp][0])
                for struct_elmt_tmp in range(1,len(diff_percentage[level_tmp])):
                    bilan_html+="<p>"+tab+diff_percentage[level_tmp][struct_elmt_tmp]+"</p>"
    else:
        bilan_html+="<p>!!Aucun niveau n'est défini => Pas d'analyse par niveau !!!</p>"    
    bilan_html += saut_de_ligne + saut_de_ligne
    numero_de_chapitre+=1
    bilan_html+='<H2 align="left"> {} - Analyse par type d\'éléments structuraux</H2>'.format(numero_de_chapitre)
    # poids_des_elements={"Unités":0.5,"Niveaux":0.1,"Fondations":0,"Murs":5,"Dalles":1,"Poutres":0,"Poteaux":0}
    for element_name in poids_des_elements:
        if poids_des_elements[element_name]>0:
            numero_de_sous_chapitre+=1
            bilan_html+='<H3 align="left"> {}.{} - {} </H3>'.format(numero_de_chapitre,numero_de_sous_chapitre,element_name)
            for sous_elmt_tmp in dico_evaluation_student_tmp.elmt_eval[element_name]:
                bilan_html+="<p><b>{}</b> : {}%".format(sous_elmt_tmp,dico_evaluation_student_tmp.elmt_eval[element_name][sous_elmt_tmp])+ print_comparison_bar(dico_evaluation_student_tmp.elmt_eval[element_name][sous_elmt_tmp])+"</p>"
            bilan_html += saut_de_ligne
    #### BILAN FINAL ####
    bilan_html += saut_de_ligne
    numero_de_chapitre+=1
    bilan_html+='<H2 align="left"> {} - Bilan</H2>'.format(numero_de_chapitre)
    for struct_elmt in poids_des_elements:
        if poids_des_elements[struct_elmt] > 0.:
            bilan_html+="<p><b>{}</b> : {}% ".format(struct_elmt,dico_evaluation_student_tmp.bilan_eval[struct_elmt])+print_comparison_bar(dico_evaluation_student_tmp.bilan_eval[struct_elmt],5)+"</p>"
    note_finale = dico_evaluation_student_tmp.bilan_eval["Note finale"]
    dico_evaluation_student_tmp.note_finale = note_finale
    bilan_html+="<p><b>=> Note finale</b></p> : {}%".format(note_finale)+print_comparison_bar(note_finale)+print_encouragement(note_finale)+"</p>"
    pdf_bilan_IFC.add_page()
    pdf_bilan_IFC.write_html(bilan_html)
    pdf_bilan_IFC.output('Analyse_comparative_du_fichier_IFC_par_type_delements_{}.pdf'.format(dico_structure_student_tmp.file.name),'F')
    return dico_evaluation_student_tmp

def computing_final_note(elmt_eval,elmt_name):
    note_finale = 0.
    poids_total = 0.
    for struct_elmt in elmt_eval :
        note_finale += elmt_eval[struct_elmt]*poids_des_sous_elements[elmt_name][struct_elmt]
        poids_total += poids_des_sous_elements[elmt_name][struct_elmt]
    if poids_total > 0:
        note_finale = note_finale/poids_total
    return note_finale

# Évaluation d'un fichier IFC par COMPARAISON avec un fichier IFC référence
def evaluate_from_a_comparison(dico_structure_ref_tmp,dico_structure_student_tmp):
    bilan_eval = {}
    # Evaluation du fichier
    # poids_des_sous_elements_Fichier={"Unité de longueur":1,"Unité de surface":1,"Unité de volume":1}
    file_eval = {}
    for sous_elem_tmp in poids_des_sous_elements_Fichier:
        file_eval[sous_elem_tmp] = 0.
    if (dico_structure_student_tmp.file.unit_of_length == dico_structure_ref_tmp.file.unit_of_length):
        file_eval["Unité de longueur"]=100.
    else:
        file_eval["Unité de longueur"]=0.
    if (dico_structure_student_tmp.file.unit_of_area == dico_structure_ref_tmp.file.unit_of_area):
        file_eval["Unité de surface"]=100.
    else:
        file_eval["Unité de surface"]=0.
    if (dico_structure_student_tmp.file.unit_of_volume == dico_structure_ref_tmp.file.unit_of_volume):
        file_eval["Unité de volume"]=100.
    else:
        file_eval["Unité de volume"]=0.
    file_eval["Note finale"]=computing_final_note(file_eval,"Unités")
    bilan_eval['Unités']=file_eval["Note finale"]
    # Evaluation de la définition des NIVEAUX
    # poids_des_sous_elements_Niveau={"Nombre de niveau":1,"Altitude des niveaux":4}
    storeys_eval={}
    storeys_eval["Nombre"] = relative_percentage(dico_structure_student_tmp.storeys_nb,dico_structure_ref_tmp.storeys_nb)
    nb_altitudes_correctes = 0.
    for i_storey_ref in dico_structure_ref_tmp.storeys:
        for i_storey_student in dico_structure_student_tmp.storeys:
            if relative_percentage(dico_structure_student_tmp.storeys[i_storey_student].altitude,dico_structure_ref_tmp.storeys[i_storey_ref].altitude) > 99.9999999999:
                nb_altitudes_correctes += 1
    storeys_eval["Altitudes"] = round(nb_altitudes_correctes/dico_structure_ref_tmp.storeys_nb*100,1)                
    storeys_eval["Note finale"] = computing_final_note(storeys_eval,"Niveaux")
    bilan_eval['Niveaux']=storeys_eval["Note finale"]
    # Evaluation des FONDATIONS
    footings_eval={}
    footings_eval["Note finale"] = computing_final_note(footings_eval,"Fondations")
    bilan_eval["Fondations"]=footings_eval["Note finale"]
    # Evaluation des MURS
    # poids_des_sous_elements_Mur={"Nombre":1,"Surf_banches":1,"Volume":1}
    walls_eval = {}
    walls_eval["Nombre"] = relative_percentage(dico_structure_student_tmp.walls_nb,dico_structure_ref_tmp.walls_nb)
    walls_eval["Surf_banches"] = relative_percentage(dico_structure_student_tmp.walls_surf_banche,dico_structure_ref_tmp.walls_surf_banche)
    walls_eval["Volume"] = relative_percentage(dico_structure_student_tmp.walls_gross_volume,dico_structure_ref_tmp.walls_gross_volume)
    walls_eval["Note finale"] = computing_final_note(walls_eval,"Murs")
    bilan_eval["Murs"]=walls_eval["Note finale"]
    # Evaluation des POTEAUX
    columns_eval = {}
    columns_eval["Note finale"] = computing_final_note(columns_eval,"Poteaux")
    bilan_eval["Poteaux"]=columns_eval["Note finale"]
    # Evaluation des POUTRES
    beams_eval = {}
    beams_eval["Note finale"] = computing_final_note(beams_eval,"Poutres")
    bilan_eval["Poutres"]=beams_eval["Note finale"]
    # Evaluation des DALLES
    # poids_des_sous_elements_Dalle={"Nb":1,"Surface":1,"Volume":1,"Niveau":0.5}
    slabs_eval = {}
    slabs_eval["Note finale"] = computing_final_note(slabs_eval,"Dalles")
    bilan_eval["Dalles"]=slabs_eval["Note finale"]
    # Calcul de la note final
    note_finale = 0.
    poids_total = 0.
    for struct_elmt in poids_des_elements:
        poids_total += poids_des_elements[struct_elmt]
        if poids_des_elements[struct_elmt] > 0:
            note_finale += poids_des_elements[struct_elmt]*bilan_eval[struct_elmt]
    note_finale = round(note_finale/poids_total,1)
    bilan_eval["Note finale"] = note_finale
  
    Evaluation_tmp = Evaluation(dico_structure_student_tmp.file.name,dico_structure_student_tmp.file.user,
                                file_eval,storeys_eval,walls_eval,slabs_eval,beams_eval,columns_eval,footings_eval,bilan_eval)
    return Evaluation_tmp


##########################################################
########### Corps du programme               #############
##########################################################

# Création d'un dictionnaire qui regroupera les données des fichiers IFC lus
dico_Structures = {}
# Données du fichier Référence
print("Analyse du fichier référence : {}".format(nom_du_fichier_IFC_Reference))
dico_Structures[nom_du_fichier_IFC_Reference] = get_structure_from_IFC(nom_du_fichier_IFC_Reference)
# Génération d'un fichier pdf à partir des éléments structurels d'un fichier IFC
generate_pdf_from_one_structure(dico_Structures[nom_du_fichier_IFC_Reference])
# Lecture et analyse des fichiers du dossier étudiants
dico_Evaluations = {}
for element in os.listdir(dossier_fichiers_etudiants):
    if not ((dossier_fichiers_etudiants[-1] == "/") or (dossier_fichiers_etudiants[-1] == "\\")):
        dossier_fichiers_etudiants += "/"
    if element.endswith('.ifc'):
        print("Analyse du fichier {}".format(element))
        dico_Structures[element] = get_structure_from_IFC(element,dossier_fichiers_etudiants)
        dico_Evaluations[element] = evaluate_from_a_comparison(dico_Structures[nom_du_fichier_IFC_Reference],dico_Structures[element])
        dico_Structures[element] = generate_pdf_for_comparison(dico_Structures[nom_du_fichier_IFC_Reference],dico_Structures[element],dico_Evaluations[element])
#--------------------------------------------------------------------------------
# Sortie d'un pdf récapitulatif des éléments de structures du projet de base = corrige prof

# Pour chaque fichier étudiant
# => Lecture du fichier étudiant
# => Sortie d'un pdf récapitulatif des différences avec le pdf prof
# => Sortie d'un IFC mettant en évidence les éléments posant problème => éléments corrects à mettre en transparence

#--------------------------------------------------------------------------------
# Récupération des éléments à vérifier
#dico_elements_to_check = {}
#dico_elements_to_check = get_specific_elements(dico_all_elements,elements_a_verifier)

#--------------------------------------------------------------------------------
# Récupération de l'ensemble des éléments IFC Building Elements

#dico_ifc_building_elements = {}
#dico_ifc_building_elements = get_specific_elements(dico_elements,IFC_Building_Elements)
